unit mailnews;  { multi-package compatibility routines }

{

Russell Schulz - russell@alpha3.ersys.edmonton.ab.ca (950124)

Copyright 1995 Russell Schulz

this code is not in the Public Domain

permission is granted to use these routines in any application regardless
of commercial status as long as the author of these routines assumes no
liability for any damages whatsoever for any reason.  have fun.

}

{
version of this unit: 1ish waffle
  (numeric indicates completeness of routines)
  (text indicates packages with complete support for those routines)
}

{
possible $define's

  debug      - debugging info
  mailgroups - directories for groups of the form
               <uucpname>.mail.<user>[.<folder>...]
               derive from directory for <uucpname>.mail

  waffle
  uupc
  fsuucp
  mksuucp
}

{
routines specific to one package must begin with package_
(e.g., waffle_forumsval, uupc_foo, fsuucp_bar, mksuucp_baz)
and must be inside an $ifdef for that package
}

{$define debug}
{$undef debug}

{$undef mailgroups}
{$define mailgroups}

{$undef waffle}
{$define waffle}

{$define uupc}
{$undef uupc}

{$define fsuucp}
{$undef fsuucp}

{$define mksuucp}
{$undef mksuucp}

interface

uses dos,genericf;

var
  mniface: string;  {interface}
  mnver: string;    {version}
  mnsub: string;    {sub-version}

{$ifdef mailgroups}
  mailprefix: string;
{$endif}

{$ifdef waffle}
  forumsetl: string;
{$endif}

{$ifdef uupc}
{$endif}

{$ifdef fsuucp}
{$endif}

{$ifdef mksuucp}
{$endif}



procedure init;
procedure shutdown;

procedure splitaddress(address: string; var site,remoteuser: string);
function sitename(s: string): string;
function basesitename(s: string): string;
function remoteuser(s: string): string;
function getheaderslow(infn: string; fieldname: string): string;

function dosfn(unixfn: string; site: string): string;
function unixfn(dosfn: string; site: string): string;
function pathizegroup(group: string; eleven: boolean): string;
function newseqstring: string;

function getmainconfig(tag: string): string;
function tagtopackage(tag: string): string;

function getuserdir(userid: string): string;
function getgroupdir(group: string): string;
function getspooldir(site: string): string;
function ismoderated(group: string): boolean;

function expandmail(address: string; userid: string): string;


implementation

type
  mnifacet=(unknown,waffle,uupc,fsuucp,mksuucp);

var
  theiface: mnifacet;
  mainconfigfn: string;
  mainconfigf: text;

{waffle-specific code}

{$ifdef waffle}
procedure waffle_forumsval(var result: string; tag: string; forumline: string);

var
  at: integer;

begin
  at := pos(tag,forumline);
  if at<>0 then
    begin
      result := trim(ltrim(copy(forumline,at+length(tag),255)));
      result := unquote(getfirstw(result));
    end;
end;
{$endif}

{uupc-specific code}
{$ifdef uupc}
{$endif}

{fsuucp-specific code}
{$ifdef fsuucp}
{$endif}

{mksuucp-specific code}
{$ifdef mksuucp}
{$endif}


{---code common to all---}

procedure splitaddress;

var
  atbang: integer;
  atpercent: integer;
  atat: integer;
  result: string;
  work: string;
  atdot: integer;

begin
  site := address;
  remoteuser := address;

  atbang := pos('!',address);
  atpercent := pos('%',address);
  atat := pos('@',address);
  if atbang>0 then
    begin
      work := address;
      while atbang>0 do
        begin
          site := copy(work,1,atbang-1);
          work := copy(work,atbang+1,255);
          atbang := pos('!',work);
        end;
      
    end
  else if atpercent>0 then
    begin
      result := copy(address,atpercent+1,255);
      atat := pos('@',result);
      if atat>0 then
        result := copy(result,1,atat-1);
    end
  else if atat>0 then
    begin
      result := copy(address,atat+1,255);
    end;
end;

function sitename;

var
  result: string;
  dummys: string;

begin
  splitaddress(s,result,dummys);
  sitename := result;
end;

function basesitename;

var
  result: string;
  atdot: integer;

begin
  result := sitename(s);
  atdot := pos('.',result);
  if atdot>0 then
    result := copy(result,1,atdot-1);
  basesitename := result;
end;

function remoteuser;

var
  result: string;
  dummys: string;

begin
  splitaddress(s,dummys,result);
  remoteuser := result;
end;

function getheaderslow;

{doesn't handle continuation lines}

const
  bufsz=2048;

var
  result: string;
  inf: file;
  lowerfieldname: string;
  done: boolean;
  s: string;
  buf: array[1..bufsz] of char;
  bytesread: integer;
  offset: integer;
  i: integer;
  matched: boolean;
  lowch: char;

begin
  result := '';
  assign(inf,infn);
  reset(inf,1);
  blockread(inf,buf,bufsz-1,bytesread);
  close(inf);

  lowerfieldname := lower(fieldname);

{fill rest of buffer, move everything down one, and set first element to LF}

  for offset := bufsz downto bytesread+2 do
    buf[offset] := chr(10);
  for offset := bytesread+1 downto 2 do
    buf[offset] := buf[offset-1];

  buf[1] := chr(10);

  offset := 1;

  done := false;
  while not done do
    begin
      if offset>=bufsz then
        done := true
      else if buf[offset]<>chr(10) then
        inc(offset)
      else if (buf[offset+1]=chr(10)) or (buf[offset+1]=chr(13)) then
        done := true
      else
        begin
          matched := true;
          for i := 1 to length(fieldname) do
            if i+offset>bufsz then
              matched := false
            else
              begin
                lowch := lowcase(buf[i+offset]);
                matched := matched and (lowch=lowerfieldname[i]);
              end;
          if matched then
            begin
              matched := false;
              inc(offset,length(fieldname));
              for i := 1 to 255 do
                begin
                  if not matched and (offset+i<=bufsz) then
                    if (buf[offset+i]=chr(13)) or (buf[offset+i]=chr(10)) then
                      matched := true
                    else
                      result := result+buf[offset+i];
                end;
              result := ltrim(result);
              done := true;
            end;
          inc(offset);
        end;
    end;
  getheaderslow := result;
end;



function dosfn;

var
  result: string;
  i: integer;
  nonsitepart: string;
{$ifdef waffle}
  mungmult: byte;
  mungval: byte;
{$endif}

begin
  result := '';

  nonsitepart := '';
  for i := 1 to length(site) do
    if copy(unixfn,3,i)=copy(site,1,i) then
      nonsitepart := copy(unixfn,3+i,255);

  case theiface of

    waffle:
      begin
{$ifdef waffle}
        mungval := 0;
        mungmult := 1;

{use only the last 5 chars, as per waffle docs}
        for i := 1 to min(length(nonsitepart),5) do
          begin
            if islower(nonsitepart[length(nonsitepart)-i+1]) then
              mungval := mungval+mungmult;
            mungmult := mungmult+mungmult;
          end;
        if mungval>9 then
          mungval := mungval+ord('A')-10
        else
          mungval := mungval+ord('0');
        result := chr(mungval)+nonsitepart+'.'+copy(unixfn,1,1);
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  dosfn := result;
end;

function unixfn;

var
  result: string;
  i: integer;
  nonsitepart: string;
{$ifdef waffle}
  mungmult: byte;
  mungval: byte;
{$endif}

begin
  result := '';

  case theiface of

    waffle:
      begin
{$ifdef waffle}

{grab everything but the leading digit, and the ending .letter}

        nonsitepart := upper(copy(dosfn,2,length(dosfn)-3));

        {now adjust capitalization for letters which should be lowercase}

        mungval := ord(upcase(dosfn[1]));
        if mungval>ord('9') then
          mungval := mungval-ord('A')+10
        else
          mungval := mungval-ord('0');

        for i := length(nonsitepart) downto 1 do
          begin
            if odd(mungval) then
              nonsitepart[i] := lowcase(nonsitepart[i]);
            mungval := mungval div 2;
          end;

        result := copy(dosfn,length(dosfn),1)+'.'+site+
         nonsitepart;
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  unixfn := result;
end;

function pathizegroup;

var
  result: string;
  mangledgroup: string;
  component: string;

begin
  result := '';
  mangledgroup := crepl(group,'.',' ');
  while mangledgroup<>'' do
    begin
      component := chopfirstw(mangledgroup);
      if length(component)<8 then
        result := result+component
      else if eleven then
        result := result+copy(component,1,8)+'.'+
         copy(component, max(9,length(component)-2), 3)
      else
        result := result+copy(component,1,8);
      if mangledgroup<>'' then
        result := result+'\';
    end;
  pathizegroup := result;
end;

function newseqstring;

var
  result: string;
{$ifdef waffle}
  seqf: text;
  seqfn: string;
{$endif}

begin
  result := '0042';
  case theiface of

    waffle:
      begin
{$ifdef waffle}

{uufree hack}

        if copy(mnver,1,1)='0' then
          seqfn := getmainconfig('waffle')+'\system\'+'seqf'
        else if mnver='1.64' then
          seqfn := getmainconfig('waffle')+'\system\'+'seqf'
        else
          seqfn := getmainconfig('waffle')+'\uucp\'+'sequence';
        assign(seqf,seqfn);
        {$I-}
        reset(seqf);
        {$I+}
        if ioresult=0 then
          begin
            readln(seqf,result);
            result := trim(result);
            close(seqf);
          end;
        rewrite(seqf);
        writeln(seqf,integertozstring(atoi(result)+1,4));
        close(seqf);
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}
  newseqstring := result;
end;

function mappedgetmainconfig(tag: string): string;

var
  result: string;
{$ifdef waffle}
  s: string;
  foundtag: string;
{$endif}

begin
  result := '';
  case theiface of

    waffle:
      begin
{$ifdef waffle}
        reset(mainconfigf);
        while (result='') and not eof(mainconfigf) do
          begin
            readln(mainconfigf,s);
            if s<>'' then
              if copy(s,1,1)<>'#' then
                begin
                  foundtag := trim(ltrim(copy(s,1,pos(':',s)-1)));
                  if foundtag=tag then
                    begin
                      result := trim(ltrim(copy(s,pos(':',s)+1,255)));
                    end;
                end;
          end;
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  mappedgetmainconfig := result;
end;

function tagtopackage;

var
  result: string;

begin
  result := tag;
  tagtopackage := result;
end;

function getmainconfig;

begin
  getmainconfig := mappedgetmainconfig(tagtopackage(tag));
end;

function getuserdir;

var
  result: string;

begin
  result := '';

  case theiface of

    waffle:
      begin
{$ifdef waffle}
        result := getmainconfig('user')+'\'+userid;
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  getuserdir := unslash(result);
end;

function getgroupdir;

var
  result: string;
  s: string;

{$ifdef mailgroups}
  partialgroup: string;
{$endif}

{$ifdef waffle}
  mungedforumsetl: string;
  oneforum: string;
  activefn: string;
  activef: text;
  defaultdir: string;
  found: boolean;
  spaceat: integer;
{$endif}

begin
  result := '';

  case theiface of

    waffle:
      begin
{$ifdef waffle}
        mungedforumsetl := forumsetl+' ';
        while (result='') and (mungedforumsetl<>' ') do
          begin
            spaceat := pos(' ',mungedforumsetl);
            oneforum := trim(copy(mungedforumsetl,1,spaceat));
            mungedforumsetl :=
             ltrim(trim(copy(mungedforumsetl,spaceat,255)))+' ';

            activefn := unslash(getmainconfig('waffle'))+
             '\'+'system'+'\'+oneforum;
            assign(activef,activefn);
            {$I-}
            reset(activef);
            {$I+}

            if ioresult=0 then
              begin

                defaultdir := '';
                found := false;

                while not eof(activef) and not found do
                  begin
                    readln(activef,s);
                    s := expand(s);
                    if copy(ltrim(s)+' ',1,8)='DEFAULT ' then
                      waffle_forumsval(defaultdir,'/dir=',s);
                    if copy(ltrim(s)+' ',1,length(group)+1)=group+' ' then
                      begin
                        found := true;
                        waffle_forumsval(result,'/dir=',s);
                      end;
                  end;

                close(activef);
                if found then
                  if result='' then
                    if defaultdir<>'' then
                      begin
                        defaultdir := withbackslash(unslash(defaultdir));
                        result := defaultdir+crepl(group,'.','\');
                      end;
              end;
          end;
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  result := unslash(result);

{$ifdef mailgroups}
  if result='' then
    begin
      if copy(group,1,length(mailprefix)+1)=mailprefix+'.' then
        begin
          partialgroup := copy(group,1,length(mailprefix));
          result := getgroupdir(partialgroup);
          if result<>'' then
            begin
              result :=
               result+crepl(copy(group,length(mailprefix)+1,255),'.','\');
            end;
        end;
    end;
{$endif}

{should all illegal chars get changed to `_' here?}

  getgroupdir := result;
end;

function getspooldir: string;

var
  result: string;

begin
  result := '';

  case theiface of

    waffle:
      begin
{$ifdef waffle}
        result := unslash(getmainconfig('spool'))+'\'+site;
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  getspooldir := result;
end;

function expandmail;

var
  result: string;
{$ifdef waffle}
  newaddressfn: string;
  newaddressf: text;
  changed: boolean;
  s: string;
{$endif}

begin
  result := address;

  case theiface of

    waffle:
      begin
{$ifdef waffle}
        changed := false;
        if (pos('@',address)=0) and (pos('!',address)=0) then
          begin
            newaddressfn := getmainconfig('waffle')+'\system\'+'aliases';

            assign(newaddressf,newaddressfn);
            {$I-}
            reset(newaddressf);
            {$I-}

            if ioresult=0 then
              begin
                while not changed and not eof(newaddressf) do
                  begin
                    readln(newaddressf,s);
                    if chopfirstw(s)=address then
                      begin
                        changed := true;
                        result := s;
                      end;
                  end;
                close(newaddressf);
              end;

            if not changed and (userid<>'') then
              begin
                newaddressfn := getuserdir(userid)+'\'+'aliases';

                assign(newaddressf,newaddressfn);
                {$I-}
                reset(newaddressf);
                {$I+}

                if ioresult=0 then
                  begin
                    while not changed and not eof(newaddressf) do
                      begin
                        readln(newaddressf,s);
                        if chopfirstw(s)=address then
                          begin
                            changed := true;
                            result := s;
                          end;
                      end;
                    close(newaddressf);
                  end;
              end;

            if not changed then
              begin

{make sure no chance of security hole - no .. or \ or / or : in address}

{don't need to make sure it's not a device - last part of name is always}
{the string 'forward'}

               if (pos('/',address)=0) and (pos(':',address)=0) and
                (pos('\',address)=0) and (pos('..',address)=0) then
                 begin
                   newaddressfn := getuserdir(address)+'\'+'forward';

                   assign(newaddressf,newaddressfn);
                   {$I-}
                   reset(newaddressf);
                   {$I+}

                   if ioresult=0 then
                     begin
                       if not eof(newaddressf) then
                         begin
                           changed := true;
                           readln(newaddressf,result);
                         end;
                       close(newaddressf);
                     end;
                 end;
              end;
          end;
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  expandmail := result;
end;

function ismoderated;

var
  result: boolean;

begin
  result := false;

  case theiface of

    waffle:
      begin
{$ifdef waffle}
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  ismoderated := result;
end;




{

template!

}

function getfoo: string;

var
  result: string;

begin
  result := '';

  case theiface of

    waffle:
      begin
{$ifdef waffle}
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end;

    end; {case}

  getfoo := result;
end;

{---initialization stuff---}

procedure detect;

begin
  if getenv('WAFFLE')<>'' then
    begin
      mniface := 'waffle';
      theiface := waffle;
    end;
  if getenv('STATIC')<>'' then
    begin
      mniface := 'waffle';
      theiface := waffle;
    end;
end;

procedure init;

begin
  if mniface='' then
    detect;

  if mniface='waffle' then
    begin
      theiface := waffle;
{$ifdef waffle}
{$ifdef debug}
      writeln('waffle support initiated');
{$endif}
      mainconfigfn := getenv('WAFFLE');
      if mainconfigfn='' then
        mainconfigfn := getenv('STATIC');
      assign(mainconfigf,mainconfigfn);
      {$I-}
      reset(mainconfigf);
      {$I+}
      if ioresult<>0 then
        begin
          writeln('could not open ',mainconfigfn);
        end;

      if mnver='' then
        mnver := getmainconfig('version');

{$ifdef debug}
      writeln('version: ',mnver);
{$endif}

{read in variables that might likely be overwritten with command-line}
{but need to be accessed by mailnews local routines (like getgroupdir)}

      forumsetl := getmainconfig('forums');

{$else}
      writeln(mniface,' support not compiled in');
{$endif}
    end

  else if mniface='uupc' then
    begin
      theiface := uupc;
{$ifdef uupc}
{$ifdef debug}
      writeln('uupc support initiated');
{$endif}
{$else}
      writeln(mniface,' support not compiled in');
{$endif}
    end

  else if mniface='fsuucp' then
    begin
      theiface := fsuucp;
{$ifdef fsuucp}
{$ifdef debug}
      writeln('fsuucp support initiated');
{$endif}
{$else}
      writeln(mniface,' support not compiled in');
{$endif}
    end

  else if mniface='mksuucp' then
    begin
      theiface := mksuucp;
{$ifdef mksuucp}
{$ifdef debug}
      writeln('mksuucp support initiated');
{$endif}
{$else}
      writeln(mniface,' support not compiled in');
{$endif}
    end

  else
    begin
      writeln('unknown interface: ',mniface);
    end;
end;

procedure shutdown;

begin
  case theiface of

    waffle:
      begin
{$ifdef waffle}
        close(mainconfigf);
{$endif}
      end;

    uupc:
      begin
{$ifdef uupc}
{$endif}
      end;

    fsuucp:
      begin
{$ifdef fsuucp}
{$endif}
      end;

    mksuucp:
      begin
{$ifdef mksuucp}
{$endif}
      end

    else
      begin
        writeln('did not know how to shut down');
      end;

    end; {case}

end;

begin
  mniface := 'unknown';
  theiface := unknown;
  mnver := '';
  mnsub := '';
end.
