program makenov;  {create/update News OVerview files}

{
Russell Schulz - russell@alpha3.ersys.edmonton.ab.ca (950327)

Copyright 1995 Russell Schulz

this code is not in the Public Domain

permission is granted to use these routines in any application regardless
of commercial status as long as the author of these routines assumes no
liability for any damages whatsoever for any reason.  have fun.
}

uses dos,genericf,mailnews,rdheader;

const
  longintsinanode=256;

type
  longintarrayt=array[1..longintsinanode] of longint;
  longintnodeptr=^longintnode;

  longintnode=
    record
      longintarray: longintarrayt;
      next: longintnodeptr;
    end;

  destinationt=(directory,group);

var
  currdir: string;
  currgroup: string;

  articlefilenamepattern: string;
  overviewfn: string;
  head: longintnodeptr;

  numexisting: integer;
  numnew: integer;

  verbose: boolean;
  emptyit: boolean;

  destination: destinationt;
  firstdestination: integer;

procedure usage;

begin
  writeln('makenov -- makes News OVerview database');
  writeln;
  writeln('version 2');
  writeln;
  writeln('usage:  makenov [options] destination');
  writeln;
  writeln('options:');
  writeln('  -v verbose');
  writeln('  -e empty out overview first if it exists');
  writeln;
  writeln('destination is one of:');
  writeln('  -d directory [directory...]');
  writeln('  -g group [group...]');
  writeln;
  writeln('russell@alpha3.ersys.edmonton.ab.ca (950327)');
  halt(1);
end;

procedure deletefile(fn: string);

var
  f: file;

begin
  assign(f,fn);
  erase(f);
end;

procedure initialize;

var
  currparami: integer;
  currparams: string;

begin
  if paramcount=0 then
    usage;

  if paramstr(1)='-?' then
    usage;

  currdir := '.';
  currgroup := '';
  articlefilenamepattern := '*';
  overviewfn := 'overview';

  verbose := false;
  emptyit := false;

  firstdestination := 0;

  currparami := 1;

  while currparami<=paramcount do
    begin
      currparams := paramstr(currparami);

      if currparams='-v' then
        verbose := true
      else if currparams='-e' then
        emptyit := true
      else if currparams='-d' then
        begin
          if currparami=paramcount then
            begin
              writeln('-d requires an argument');
              usage;
            end;

          firstdestination := currparami+1;
          currparami := paramcount+1;
          destination := directory;
        end
      else if currparams='-g' then
        begin
          if currparami=paramcount then
            begin
              writeln('-g requires an argument');
              usage;
            end;

          firstdestination := currparami+1;
          currparami := paramcount+1;
          destination := group;
        end
      else
        begin
          writeln('unknown option: ',currparams);
          usage;
        end;

      inc(currparami);
    end;

  if firstdestination=0 then
    begin
      writeln('must use -d or -g');
      usage;
    end;

  if destination=group then
    begin
      mniface := '';
      mailnews.init;

      if mniface='' then
        begin
          writeln('could not initialize mail/news interface');
          writeln('make sure all environment variables are set');
          usage;
        end;
  
      mailnews.mailprefix := '--not-applicable--';
    end;
end;

function asdir(agroupordir: string): string;

var
  result: string;

begin
  result := agroupordir;

  if destination=group then
    begin
      result := getgroupdir(agroupordir);
      if result='' then
        begin
          writeln('cannot find directory for group ',agroupordir);
          usage;
        end;
    end;

  asdir := result;
end;

function firstdirectory: string;

var
  result: string;

begin
  result := asdir(paramstr(firstdestination));
  inc(firstdestination);
  firstdirectory := result;
end;

function nextdirectory: string;

var
  result: string;

begin
  result := '';
  if firstdestination<=paramcount then
    begin
      result := asdir(paramstr(firstdestination));
      inc(firstdestination);
    end;

  nextdirectory := result;
end;

function alreadyinoverview(artnum: longint): boolean;

var
  result: boolean;
  anode: longintnodeptr;
  arrayind: integer;

begin
  result := false;

  anode := head;
  while not result and (anode<>nil) do
    begin
      for arrayind := 1 to longintsinanode do
        if artnum=anode^.longintarray[arrayind] then
          result := true;
      anode := anode^.next;
    end;

  alreadyinoverview := result;
end;

procedure markaspresent(artnum: longint);

procedure markaspresentnodes
 (var anode: longintnodeptr; comesfrom: longintnodeptr; artnum: longint);

var
  currnode: longintnodeptr;
  prevnode: longintnodeptr;
  marked: boolean;
  arrayind: integer;

begin
  marked := false;

  currnode := anode;
  prevnode := comesfrom;

  while not marked do
    begin
      if currnode=nil then
        begin
          if verbose then
            writeln('creating a new node...');

          new(currnode);
          if anode=nil then
            anode := currnode;

          currnode^.next := nil;
          for arrayind := 1 to longintsinanode do
            currnode^.longintarray[arrayind] := -1;

          if prevnode<>nil then
            prevnode^.next := currnode;
        end;

        for arrayind := 1 to longintsinanode do
          if not marked then
            if currnode^.longintarray[arrayind]<0 then
              begin
                marked := true;
                currnode^.longintarray[arrayind] := artnum;
              end;

        if not marked then
          begin
            prevnode := currnode;
            currnode := currnode^.next;
          end;
    end;
end;

begin
  markaspresentnodes(head,nil,artnum);
end;

procedure initoverview(adir: string; var novfn: string);

var
  novf: text;
  novline: string;
  novartfn: string;

begin
  novfn := adir+'\'+overviewfn;

  head := nil;

  numexisting := 0;
  numnew := 0;

  if not fexists(novfn) then
    writeln(novfn,' not found')
  else if emptyit then
    begin
      deletefile(novfn);
      writeln(novfn,' emptied');
    end
  else
    begin
      assign(novf,novfn);
      reset(novf);

      while not eof(novf) do
        begin
          readln(novf,novline);
          novartfn := getfirstw(novline);

          if verbose then
            writeln(novartfn,' already in overview file')
          else
            write('.');

          inc(numexisting);

          markaspresent(atol(novartfn));
        end;

      close(novf);
    end;

  if not verbose then
    if numexisting<>0 then
      writeln;

  writeln('existing: ',numexisting);
end;

procedure freehead;

var
  currnode: longintnodeptr;
  nextnode: longintnodeptr;

begin
  currnode := head;

  while currnode<>nil do
    begin
      nextnode := currnode^.next;
      currnode^.next := nil;
      dispose(currnode);
      currnode := nextnode;
    end;

  head := nil;
end;

function notabheader(filename: string; tag: string): string;

var
  result: string;

begin
  result := getheaderline(filename,tag);
  result := crepl(result,tab,' ');
  notabheader := result;
end;

procedure process(oldadir: string);

var
  adir: string;
  fileinfo: searchrec;
  artnum: longint;
  novfn: string;
  novf: text;
  novline: string;
  novartfn: string;

begin
  adir := oldadir;

  if copy(adir,length(adir),1)='\' then
    adir := adir+'.';
  if copy(adir,length(adir),1)=':' then
    adir := adir+'.';

  writeln('directory=',adir);

  initoverview(adir,novfn);

  assign(novf,novfn);
{$I-}
  append(novf);
{$I-}
  if ioresult<>0 then
    rewrite(novf);

  if emptyit then
    rewrite(novf);

  findfirst(adir+'\'+articlefilenamepattern,archive,fileinfo);
  while doserror=0 do
    begin
      artnum := atol(fileinfo.name);
      if isdigit(fileinfo.name[1]) then
        if not alreadyinoverview(artnum) then
          begin
            if verbose then
              writeln('adding ',fileinfo.name,'...')
            else
              write('.');

            inc(numnew);

            novartfn := adir+'\'+fileinfo.name;
            writeln(novf,
             fileinfo.name,tab,
             notabheader(novartfn,'subject:'),tab,
             notabheader(novartfn,'from:'),tab,
             notabheader(novartfn,'date:'),tab,
             notabheader(novartfn,'message-id:'),tab,
             notabheader(novartfn,'references:'),tab,
             fileinfo.size,tab,
             '',tab,  {lines}
             'Newsgroups: ',notabheader(novartfn,'newsgroups:'));
          end;
      findnext(fileinfo);
    end;

  if not verbose then
    if numnew<>0 then
      writeln;

  writeln('new: ',numnew);

  close(novf);

  freehead;
end;

begin
  initialize;

  currdir := firstdirectory;
  while currdir<>'' do
    begin
      process(currdir);
      currdir := nextdirectory;
    end;
end.
